//DxClipPolygon.cpp
#include <assert.h>
#include <d3d9.h>
#include <d3dx9.h>
#include "ray.h"
#include "DxPolygon.h"
#include "Clipping.h"
#include "MeshObj.h"
#include "ObbMesh.h"
#include "DxClipPolygon.h"


//----------------------------------------------------------------------------------------------
void DxClipPolygon::SetDecalPos( float x, float y, float z )
{
	m_decalPos = D3DXVECTOR3( x, y, z );
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::SetDecalLookAt( float x, float y, float z )
{
	m_decalLookAt = D3DXVECTOR3( x, y, z );
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::SetDecalRadius( float x, float y, float z )
{
	m_decalRadius = D3DXVECTOR3( x, y, z );
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::GetViewMatrix( D3DXMATRIX& m,  const D3DXVECTOR3 &vFrom, 
								  const D3DXVECTOR3 &vView, const D3DXVECTOR3 &vUp, const D3DXVECTOR3 &vRight)
{
	// 9 muls, 9 adds
	m._11 = vRight.x;			 m._12 = vUp.x;			 m._13 = vView.x;		      m._14 = 0.0f;
	m._21 = vRight.y;			 m._22 = vUp.y;			 m._23 = vView.y;		      m._24 = 0.0f;
	m._31 = vRight.z;			 m._32 = vUp.z;			 m._33 = vView.z;		      m._34 = 0.0f;
	
	m._41 = -D3DXVec3Dot( &vFrom, &vRight);   
	m._42 = -D3DXVec3Dot( &vFrom, &vUp); 
	m._43 = -D3DXVec3Dot( &vFrom, &vView);    
	m._44 = 1.0f;			
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::GetViewMatrix( D3DXMATRIX* pMatView )
{
#if   1
	D3DXMATRIX	 matCameraView;
	//D3DXVECTOR3  cameraTarget = D3DXVECTOR3( 0, 0, 0 );
	D3DXVECTOR3  cameraTarget = m_decalLookAt;
    D3DXVECTOR3  cameraPos = m_decalPos;

	// Setup Camera View
	D3DXMatrixLookAtRH( &matCameraView, 
						&cameraPos, 
						&cameraTarget, 
						&D3DXVECTOR3( 0, 1, 0 ) );

	memcpy( pMatView, matCameraView, sizeof( D3DXMATRIX ) );
#else
	D3DXMATRIX	 matCameraView;
	D3DXVECTOR3  cameraPos = m_decalPos;

	GetViewMatrix( matCameraView, cameraPos, D3DXVECTOR3( 0, 0, 1 ), D3DXVECTOR3( 0, 1, 0 ), D3DXVECTOR3( 1, 0, 0 ) );

	D3DXMATRIX rot;
	D3DXMatrixRotationYawPitchRoll( &rot, 0, 0, D3DXToRadian( 0 ) );
	D3DXMatrixMultiply( pMatView, &rot, &matCameraView );
#endif
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::GetProjectionMatrix( D3DXMATRIX* pMatProj )
{
	D3DXMATRIX   matProj;
	D3DXMatrixOrthoRH( pMatProj, m_decalRadius.x * 2,  m_decalRadius.y * 2, 0.0f, m_decalRadius.z );  
}
//----------------------------------------------------------------------------------------------
DxClipPolygon::DxClipPolygon()
{
	m_pObbBox = NULL;
	m_decalPos = D3DXVECTOR3( 0, 0, 0 );
	m_decalLookAt = D3DXVECTOR3( 0, 0, -1 );
	m_decalRadius = D3DXVECTOR3( 1, 1, 1 );
	m_pCameraVertex = NULL;
}
//----------------------------------------------------------------------------------------------
DxClipPolygon::~DxClipPolygon()
{

}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::Cleanup()
{
	DxPolygon::Cleanup();
	if( m_pObbBox )
	{
		delete m_pObbBox;
		m_pObbBox = NULL;
	}

	if( m_pCameraVertex )
	{
		delete [] m_pCameraVertex;
		m_pCameraVertex = NULL;
		m_nCameraCount = 0;
	}
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::MakeObbBox( const D3DXMATRIX& mat )
{
	assert( m_pObbBox == NULL );
	m_pObbBox = new ObbMesh();
	m_pObbBox->SetRadius( m_decalRadius.x, m_decalRadius.y, m_decalRadius.z );
	m_pObbBox->SetTranslate( m_decalPos.x, m_decalPos.y, 0 );

	D3DXVECTOR3 angle;
	ToEulerAnglesYXZ( &angle, &mat );

	m_pObbBox->SetRotate( angle.x, angle.y, angle.z );
	m_pObbBox->SetColor( 0, 0, 1 );
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::MakeCameraLine( const D3DXMATRIX& invMat )
{
	m_cameraLine.clear();
	PointArray cameraLine;
	cameraLine.push_back( DataType::Point3( -1, -1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 0 ) );
	cameraLine.push_back( DataType::Point3( -1,  1, 0 ) );
	cameraLine.push_back( DataType::Point3( -1,  1	, 0 ) );
	cameraLine.push_back( DataType::Point3( -1, -1, 0 ) );

	cameraLine.push_back( DataType::Point3( -1, -1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 1 ) );
	cameraLine.push_back( DataType::Point3( -1,  1, 1 ) );
	cameraLine.push_back( DataType::Point3( -1,  1, 1 ) );
	cameraLine.push_back( DataType::Point3( -1, -1, 1 ) );

	cameraLine.push_back( DataType::Point3( -1, -1, 0 ) );
	cameraLine.push_back( DataType::Point3( -1, -1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1, -1, 1 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 0 ) );
	cameraLine.push_back( DataType::Point3(  1,  1, 1 ) );
	cameraLine.push_back( DataType::Point3( -1,  1, 0 ) );
	cameraLine.push_back( DataType::Point3( -1,  1, 1 ) );

	D3DXVECTOR3	invVec;
	int iCount = (int)cameraLine.size();
	for( int i = 0; i < iCount; ++i )
	{
		D3DXVECTOR3 vec = D3DXVECTOR3( cameraLine[i].x, cameraLine[i].y, cameraLine[i].z );
		D3DXVec3TransformCoord( &invVec, &vec, &invMat );
		m_cameraLine.push_back( DataType::Point3( invVec.x, invVec.y, invVec.z) );
	}
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::CopyCameraLine()
{
	//ī޶  .
	if( m_cameraLine.size() < 2 )
		return;
	int nNum = m_nCameraCount = (int)m_cameraLine.size();
	assert( m_pCameraVertex == NULL );
	m_pCameraVertex = new DataType::WireVertex [ nNum ];

	for( int i = 0; i < nNum; ++i )
	{
		m_pCameraVertex[i].x = m_cameraLine[i].x;
		m_pCameraVertex[i].y = m_cameraLine[i].y;
		m_pCameraVertex[i].z = m_cameraLine[i].z;
		m_pCameraVertex[i].color = D3DXCOLOR( 0, 0, 1, 1.0f );
	}
}
//----------------------------------------------------------------------------------------------
void DxClipPolygon::TestDecalmap()
{
	if( m_nCount == 0 )
		return;

	m_texUV.clear();
	D3DXMATRIX  matProj, matCameraView, matWorld;

	D3DXMatrixTranslation( &matWorld, 0, 0, 0 );

	GetViewMatrix( &matCameraView );
	GetProjectionMatrix( &matProj );

	D3DXMATRIXA16  mat = matWorld * matCameraView * matProj;

	D3DXVECTOR3 out;
	PointArray temp;
	for( int i = 0; i < m_nCount; ++i )
	{
		temp.push_back( DataType::Point3( m_position[i].x, m_position[i].y, m_position[i].z ) );
	}
	m_position.clear();

	//-- Ŀ  ȯ
	for( int i = 0; i < m_nCount; ++i )
	{
		D3DXVECTOR3 vec = D3DXVECTOR3( temp[i].x, temp[i].y, temp[i].z );
		D3DXVec3TransformCoord(&out, &vec, &mat);
		
		m_position.push_back( DataType::Point3( out.x, out.y, out.z ) );
	}

	ClipAxis( 1, 0, 0, -1, false );
	ClipAxis( -1, 0, 0, 1, false );
	ClipAxis( 0, 1, 0, -1, false );
	ClipAxis( 0, -1, 0, 1, false );

	//z ڸ
	ClipAxis( 0, 0, -1, 1, false );

	temp.clear();
	int iCount = (int)m_position.size();
	for( int i = 0; i < iCount; ++i )
	{
		temp.push_back( DataType::Point3( m_position[i].x, m_position[i].y, m_position[i].z ) );
	}
	m_position.clear();


	//-- Ŀ  纯ȯ
	D3DXMATRIX invMat;
	D3DXMatrixInverse( &invMat, 0, &mat );
	D3DXVECTOR3	invVec;
	for( int i = 0; i < iCount; ++i )
	{
		D3DXVECTOR3 vec = D3DXVECTOR3( temp[i].x, temp[i].y, temp[i].z );
		D3DXVec3TransformCoord( &invVec, &vec, &invMat );
		m_position.push_back( DataType::Point3( invVec.x, invVec.y, invVec.z) );

		//ؽ 
		float u = 0.5f * vec.x + 0.5f;
		float v = -0.5f * vec.y + 0.5f;
		m_texUV.push_back( DataType::Point2( u, v ) );
	}

	MakeCameraLine( invMat );

	ReComputeNormal();
	MakeObbBox( mat );
}
//------------------------------------------------------------------------------------------------
void DxClipPolygon::CopyData()
{
	int nNum = m_nCount = (int)m_position.size();
	int nNormalNum = (int)m_normal.size();
	int nUVNum = (int)m_texUV.size();

	m_pVertex = new DataType::Vertex [ nNum ];
	m_pWireVertex = new DataType::WireVertex [ nNum ];

	for( int i = 0; i < nNum; ++i )
	{
		//ְ 0.01 Ͽ Z-Fighting Ͼ ʵ Ѵ
		m_pVertex[i].x = m_position[i].x + m_normal[i].x*0.01f;
		m_pVertex[i].y = m_position[i].y + m_normal[i].y*0.01f;
		m_pVertex[i].z = m_position[i].z + m_normal[i].z*0.01f;
		
		if( nNum == nNormalNum )
		{
			m_pVertex[i].nx = m_normal[i].x;
			m_pVertex[i].ny = m_normal[i].y;
			m_pVertex[i].nz = m_normal[i].z;
		}
		else
		{
			assert( 0 && "nNum == nNormalNUm" );
		}

		if( nNum == nUVNum )
		{
			m_pVertex[i].u = m_texUV[i].x;
			m_pVertex[i].v = m_texUV[i].y;
		}
		else
		{
			assert( 0 && "nNum == nUVNum" );
		}

		m_pVertex[i].color = m_color;

		m_pWireVertex[i].x = m_position[i].x;
		m_pWireVertex[i].y = m_position[i].y;
		m_pWireVertex[i].z = m_position[i].z;
		m_pWireVertex[ i ].color = 0xff0000;
	}
}
//--------------------------------------------------------------------------------------------------------------------
void DxClipPolygon::Render( const LPDIRECT3DDEVICE9 device )
{
	if( m_pVertex == NULL )
	{
		if( m_position.empty() )
			return;

		CopyData();
		CopyCameraLine();
	}

	//  ̵
	D3DXMATRIX	matRot, matTrans, matScale;
	D3DXMatrixTranslation( &matTrans, m_translate.x, m_translate.y, m_translate.z );
	D3DXMatrixRotationYawPitchRoll( &matRot, m_rotate.y, m_rotate.x, m_rotate.z );
	D3DXMatrixScaling( &matScale, m_scale.x, m_scale.y, m_scale.z );
	D3DXMATRIXA16  matWorld = matScale * matRot * matTrans;
	device->SetTransform( D3DTS_WORLD, &matWorld );

	HRESULT hr;
	if( m_pTexture )
		device->SetTexture( 0, m_pTexture );
	// Draw the vertex list
	device->SetFVF( DataType::Type );
	device->SetRenderState( D3DRS_LIGHTING, TRUE );
	device->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pVertex, sizeof(DataType::Vertex ) );

	device->SetFVF( DataType::WireType );
	device->SetRenderState( D3DRS_LIGHTING, FALSE );
	device->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pWireVertex, sizeof(DataType::WireVertex ) );
	if( m_pCameraVertex )
		hr = device->DrawPrimitiveUP( D3DPT_LINELIST, m_nCameraCount/2, m_pCameraVertex, sizeof(DataType::WireVertex ) );
	if( m_pTexture )
		device->SetTexture( 0, NULL );

	//if( m_pObbBox )
	//	m_pObbBox->Render( device );
}

