//Polygon.cpp
#include <assert.h>
#include <d3d9.h>
#include <d3dx9.h>
#include "ray.h"
#include "DxPolygon.h"
#include "Clipping.h"
#include "MeshObj.h"

namespace DataType
{
	int Type = ( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1 );
};


namespace DataType
{
	int WireType = ( D3DFVF_XYZ | D3DFVF_DIFFUSE );
};


DxPolygon::DxPolygon()
{
	m_translate = m_rotate = D3DXVECTOR3( 0, 0, 0 );
	m_scale = D3DXVECTOR3( 1, 1, 1 );

	m_pVertex = NULL;
	m_pWireVertex = NULL;
	m_nCount = 0;
	m_color =  D3DCOLOR_XRGB(255, 255, 255);
}
//------------------------------------------------------------------------------------------------
DxPolygon::~DxPolygon()
{
	Cleanup();
}
//------------------------------------------------------------------------------------------------
void DxPolygon::Cleanup()
{
	if( m_pVertex )
	{
		delete [] m_pVertex;
		m_pVertex = NULL;
		delete [] m_pWireVertex;
		m_pWireVertex = NULL;
		m_nCount = 0;

		_RELEASE_<LPDIRECT3DTEXTURE9>( m_pTexture );
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::LoadObj( const char* szName )
{
	mesh* object = new mesh( szName);

	std::vector<mesh::Vec3>& vList = object->vList;
	std::vector<mesh::Vec3>& nList = object->nList;
	std::vector<mesh::Vec2>& tList = object->tList;
	std::vector<mesh::FACE>& faceList = object->faceList;
	int n = (int) faceList.size();

	for( int i = 0; i < n; ++i )
	{
		mesh::Vertex v0 = faceList[i].v[0];
		mesh::Vertex v1 = faceList[i].v[1];
		mesh::Vertex v2 = faceList[i].v[2];

		mesh::Vec3 vec = vList[ v0.v ]; mesh::Vec3 nom = nList[ v0.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );

		vec = vList[ v1.v ];  nom = nList[ v1.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );

		vec = vList[ v2.v ];  nom = nList[ v2.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );
	}

	if( tList.empty() == false )
	{
		for( int i = 0; i < n; ++i )
		{
			mesh::Vertex v0 = faceList[i].v[0];
			mesh::Vertex v1 = faceList[i].v[1];
			mesh::Vertex v2 = faceList[i].v[2];

			mesh::Vec2 uv = tList[ v0.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );

			uv = tList[ v1.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );

			uv = tList[ v2.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );
		}
	}
	delete object;
}
//------------------------------------------------------------------------------------------------
bool DxPolygon::SetTexture( const LPDIRECT3DDEVICE9 pDevice, const char* szFileName )
{
	HRESULT hr = D3DXCreateTextureFromFile( pDevice, szFileName, &m_pTexture );
	return SUCCEEDED( hr );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::SetVertex( DataType::Vertex* pVertex, int nNum )
{
	if( m_nCount > 0 && m_nCount != nNum )
		Cleanup();

	m_nCount = nNum;
	if( m_pVertex == NULL )
	{
		m_pVertex = new DataType::Vertex [ nNum ];
		m_pWireVertex = new DataType::WireVertex [ nNum ];
	}

	memcpy( m_pVertex, pVertex, nNum * sizeof( DataType::Vertex ) );
	for( int k = 0; k < nNum; ++k )
	{
		m_pWireVertex[k].x = pVertex[k].x;
		m_pWireVertex[k].y = pVertex[k].y;
		m_pWireVertex[k].z = pVertex[k].z;
		m_pWireVertex[k].color = 0xff0000;
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::PushBackPosition( DataType::Point3& pos )
{
	m_position.push_back( pos );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::PushBackNormal( DataType::Point3& pos )
{
	m_normal.push_back( pos );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::ReComputeNormal()
{
	if( m_pVertex )
		Cleanup();

	m_normal.clear();
	ComputePolygonNormal( this );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::ClipAxis( float nx, float ny, float nz, float basis, bool bComputeNormal )
{
	int nNum = (int)m_position.size();
	D3DXVECTOR3* vec = new D3DXVECTOR3[ nNum ];

	for( int i = 0; i < nNum; ++i )
	{
		vec[i].x = m_position[i].x;
		vec[i].y = m_position[i].y;
		vec[i].z = m_position[i].z;
	}
	m_position.clear();
	m_normal.clear();
	m_texUV.clear();

	float out[15];
	for( int i = 0; i < nNum;i += 3 )
	{
		int n = ClipTriangle( &vec[i], out, nx, ny, nz, basis );
		if( n > 0 )
		{
			assert( n <= 4 );
			if( n == 3 )
			{
				for( int k = 0; k < n*3; k += 3 )
				{
					m_position.push_back( DataType::Point3( out[k], out[k + 1], out[k + 2] ) );
				}
			}
			else if( n == 4 )
			{
				int k = 0;
				for( k = 0; k < 3*3; k += 3 )
				{
					m_position.push_back( DataType::Point3( out[k], out[k + 1], out[k + 2] ) );
				}

				int k0 = 0 * 3;
				int k1 = 2 * 3;
				int k2 = 3 * 3;

				m_position.push_back( DataType::Point3( out[k0], out[k0 + 1], out[k0 + 2] ) );	
				m_position.push_back( DataType::Point3( out[k1], out[k1 + 1], out[k1 + 2] ) );	
				m_position.push_back( DataType::Point3( out[k2], out[k2 + 1], out[k2 + 2] ) );	
			}
			else
			{
				assert( 0 && "n isn't 3 or 4" );
			}
		}
	}

	if( bComputeNormal )
		ReComputeNormal();
}
//------------------------------------------------------------------------------------------------
void DxPolygon::CopyData()
{
	int nNum = m_nCount = (int)m_position.size();
	int nNormalNum = (int)m_normal.size();
	int nUVNum = (int)m_texUV.size();

	m_pVertex = new DataType::Vertex [ nNum ];
	m_pWireVertex = new DataType::WireVertex [ nNum ];

	for( int i = 0; i < nNum; ++i )
	{
		m_pVertex[i].x = m_position[i].x;
		m_pVertex[i].y = m_position[i].y;
		m_pVertex[i].z = m_position[i].z;
		
		if( nNum == nNormalNum )
		{
			m_pVertex[i].nx = m_normal[i].x;
			m_pVertex[i].ny = m_normal[i].y;
			m_pVertex[i].nz = m_normal[i].z;
		}
		else
		{
			//TODO:  ڵ  ߰
			m_pVertex[i].nx = 1.0f;
			m_pVertex[i].ny = 1.0f;
			m_pVertex[i].nz = 1.0f;
		}

		if( nNum == nUVNum )
		{
			m_pVertex[i].u = m_texUV[i].x;
			m_pVertex[i].v = m_texUV[i].y;
		}
		else
		{
			m_pVertex[i].u = 0;
			m_pVertex[i].v = 0;			
		}

		m_pVertex[i].color = m_color;

		m_pWireVertex[i].x = m_position[i].x;
		m_pWireVertex[i].y = m_position[i].y;
		m_pWireVertex[i].z = m_position[i].z;
		m_pWireVertex[ i ].color = 0xff0000;
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::Render( const LPDIRECT3DDEVICE9 device )
{
	if( m_pVertex == NULL )
	{
		if( m_position.empty() )
			return;

		CopyData();
	}

	//  ̵
	D3DXMATRIX	matRot, matTrans, matScale;
	D3DXMatrixTranslation( &matTrans, m_translate.x, m_translate.y, m_translate.z );
	D3DXMatrixRotationYawPitchRoll( &matRot, m_rotate.y, m_rotate.x, m_rotate.z );
	D3DXMatrixScaling( &matScale, m_scale.x, m_scale.y, m_scale.z );
	D3DXMATRIXA16  matWorld = matScale * matRot * matTrans;
	device->SetTransform( D3DTS_WORLD, &matWorld );

	HRESULT hr;
	if( m_pTexture )
		device->SetTexture( 0, m_pTexture );
	// Draw the vertex list
	device->SetFVF( DataType::Type );
	device->SetRenderState( D3DRS_LIGHTING, TRUE );
	device->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pVertex, sizeof(DataType::Vertex ) );

	device->SetFVF( DataType::WireType );
	device->SetRenderState( D3DRS_LIGHTING, FALSE );
	device->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pWireVertex, sizeof(DataType::WireVertex ) );

	if( m_pTexture )
		device->SetTexture( 0, NULL );
}
//------------------------------------------------------------------------------------------------
extern   float g_gridStride;
extern HWND					g_hRenderWnd;
extern LPDIRECT3DDEVICE9    g_d3d_Device;
extern void SetupView( D3DXMATRIX* pMatView = NULL );
extern void SetupProjection( HWND hWnd,  D3DXMATRIX* pMatProj );

void DxPolygon::TestTransform( float x, float y, float z )
{
	if( m_nCount == 0 )
		return;
	
	D3DXMATRIX  matProj;
	D3DXMATRIX	matCameraView;

	SetupView( &matCameraView );
	SetupProjection( g_hRenderWnd, &matProj );

	D3DXMATRIX  matWorldViewProj;
	D3DXMATRIX  matWorldView;
	D3DXMATRIX	matWorld;
	D3DXMatrixTranslation( &matWorld, x, y, z );
	D3DXVECTOR3 out;


	for( int i = 0; i < m_nCount; ++i )
	{
		D3DXVECTOR3 vec = D3DXVECTOR3( m_position[i].x, m_position[i].y, m_position[i].z );
		
		//--  
		/*
		1
		D3DXMatrixMultiply( &matWorldView, &matWorld, &matCameraView );
		D3DXMatrixMultiply( &matWorldViewProj, &matWorldView, &matProj );
		D3DXVec3TransformCoord( &out, &vec, &matWorldViewProj );
		2
		D3DXMatrixMultiply( &matWorldView, &matWorld, &matCameraView );
		D3DXVec3TransformCoord(&vec, &vec, &matWorldView);
		D3DXVec3TransformCoord(&out, &vec, &matProj);
		*/
		//3
		D3DXMATRIXA16  mat = matWorld * matCameraView * matProj;
		D3DXVec3TransformCoord(&out, &vec, &mat);
		
		m_pVertex[i].x = out.x;
		m_pVertex[i].y = out.y;
		m_pVertex[i].z = out.z;

		//ؽ 
#if 0
		float w = 1.0f;
		float u = 0.5f * ( out.x/w ) + 0.5f;
		float v = -0.5f * ( out.y/w ) + 0.5f;

		m_pVertex[i].u = u;
		m_pVertex[i].v = v;
#endif

#if  0
		D3DXVECTOR3 backVec;
		D3DXMATRIX  backMat;  //    
		D3DXMatrixInverse( &backMat, 0, &mat );
		D3DXVec3TransformCoord( &backVec, &out, &backMat );
#endif
	}
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetTranslate()
{
	return m_translate;
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetRotate()
{
	return m_rotate;
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetScale()
{
	return m_scale;
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetTranslate( float x, float y, float z )
{
	m_translate = D3DXVECTOR3( x, y, z );
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetRotate( float x, float y, float z )
{
	m_rotate = D3DXVECTOR3( x, y, z );
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetScale( float x, float y, float z )
{
	m_scale = D3DXVECTOR3( x, y, z );
}