//DxPolygon.h
#ifndef  __DXPOLYGON_H
#define  __DXPOLYGON_H

#include <vector>

template<class T> void _RELEASE_(T t)
{
	if( t )
	{
		t->Release();
		t = 0;
	}
}

template<class T> void _DELETE_(T t)
{
	if( t )
	{
		delete t;
		t = 0;
	}
}

namespace DataType
{
	extern int Type;
	extern int WireType;

	struct Vertex
	{
		float   x, y, z; 
		float   nx, ny, nz;
		DWORD   color; 
        float      u;
        float      v;
	};

	struct WireVertex
	{
		float x, y, z; 
		int   color; 
	};

	class Point2
	{
	public:
		float x, y;
		Point2( float _x, float _y )
		{
			x = _x; y = _y; 
		}
	};

	class Point3
	{
	public:
		float x, y, z;
		Point3( float _x, float _y, float _z )
		{
			x = _x; y = _y; z = _z;
		}
	};
};

typedef std::vector< DataType::Point2 >  Point2Array;
typedef std::vector< DataType::Point3 >  PointArray;


class DxPolygon
{
public:
	DxPolygon();
	virtual ~DxPolygon();

	void	LoadObj( const char* szName );
	void	SetVertex( DataType::Vertex* pVertex, int nNum );
	virtual void	Render( const LPDIRECT3DDEVICE9 device );
	void	PushBackPosition( DataType::Point3& pos );
	void	PushBackNormal( DataType::Point3& pos );
	
	PointArray& GetPosition() { return  m_position;  }
	PointArray& GetNormal() { return  m_normal;  }
	void	SetColor( DWORD color )  
	{ 
		m_color = color;  
	}

	bool	SetTexture( const LPDIRECT3DDEVICE9 pDevice, const char* szFileName );
	void    TestTransform( float x, float y, float z );
	void	ClipAxis( float nx, float ny, float nz, float basis, bool bComputeNormal = true );

	const	D3DXVECTOR3& GetTranslate(); 
	const	D3DXVECTOR3& GetRotate();
	const	D3DXVECTOR3& GetScale();
	void	SetTranslate( float x, float y, float z );
	void	SetRotate( float x, float y, float z );
	void	SetScale( float x, float y, float z );

protected:
	virtual void	Cleanup();
	virtual void	CopyData();
	void	ReComputeNormal();

	LPDIRECT3DTEXTURE9		m_pTexture;
	DataType::Vertex*		m_pVertex;
	DataType::WireVertex*	m_pWireVertex;
	int						m_nCount;

	DWORD					m_color;
	PointArray				m_position;
	PointArray				m_normal;
	Point2Array				m_texUV;


	//̵, ȸ ũ
	D3DXVECTOR3				m_translate;
	D3DXVECTOR3				m_rotate;
	D3DXVECTOR3				m_scale;
};

#endif
